/*	LIB_MATCH_DIR_SPEC.C

Editors:
	28-AUG-1987	Joe Bryan
	Creation.


Abstract:
	Compares two directory specifications to determine if they match.  A 
	pattern specification that may contain wildcards is used to check a 
	candidate specification that cannot have wildcards.  The function 
	assumes that both directory specifications are valid VMS directory 
	specifications.

Format:	
	LIB_MATCH_DIR_SPEC	cand_str, pattern_str

Return:	
	Boolean longword: pattern_str allows cand_str as a valid match.

Arguments:
	cand_str

	VMS usage: char_string
	type: character string
	access: read only
	mechanism: by descriptor

	String compared against the pattern string.  The cand_str must be a 
	valid VMS directory specification and cannot contain any wildcard 
	operators.

	pattern_str

	VMS usage: char_string
	type: character string
	access: read only
	mechanism: by descriptor

	String used to define what is valid as a match.  The pattern_str must 
	be a valid VMS directory specification and may contain wildcard 
	operators.

Side effects:
	None.

Implicit arguments:
	None.

 */

/* -------------------------------------------------------------------------- */
/* typ_def.h */
/* definitions of data types, etc. */

#define TYPE		typedef

#define CONST		static readonly
#define VAR		static noshare

#define GDCONST		globaldef readonly
#define GRCONST		globalref

#define GDVAR		globaldef noshare
#define GRVAR		globalref

#define word		unsigned short
#define byte		unsigned char

#define TRUE	1
#define FALSE	0

/* DSCR_DEF.H	descriptor macros and definitions */

/* this file must be included after typ_def.h */

#include descrip

/* -------------------------------------------------------------------------- */
/* definition of string descriptor structure format                           */

	TYPE struct {
		word	length;		/* length data item in bytes */
		byte	type;		/* type code = DSC$K_DTYPE_T */
		byte	class;		/* class code = DSC$K_CLASS_S */
		char	*string;	/* address of string */
	} DSCR;

/* -------------------------------------------------------------------------- */
/* declarations of descriptor functions                                       */

	/* The following two functions are for passing strings to	*/
	/* routines that require the descriptor passing mechanism	*/
	/* the first function is for nul-terminated data strings,	*/
	/* while the second function is for arbitrary length data	*/
	/* buffers.  Use the first function for input to a routine	*/
	/* and use the second function for output from a routine.	*/

	DSCR	*mkdscr		(/*	*DSCR, *char		*/);
	DSCR	*mkdscrl	(/*	*DSCR, *char, long	*/);

	/* The following two functions are for accessing strings passed */
	/* into a C function as either input or output strings.  Due to */
	/* the limitations of nul-terminated strings, strings passed by */
	/* descriptor must be copied into and out of nul-terminated buffers */
	/* if the standard string processing functions are to be used.	*/

	DSCR	*cpyto_dscr	(/*	*DSCR, *char		*/);
	char	*cpyfm_dscr	(/*	*char, *DSCR, long	*/);

/* -------------------------------------------------------------------------- */
/* descriptor for single dimensional array of fixed length strings            */

	TYPE struct {
		word	length;		/* length of each array element */
		byte	dtype;		/* set to DSC$K_DTYPE_T */
		byte	class;		/* set to DSC$K_CLASS_A */
		long	base;		/* address of first element */
		word	fill;
		byte	flags;		/* set to hex D0 */
		byte	dim_cnt;	/* set to 1 */
		long	arsize;		/* size of array = length * num_elem */
		long	addr0;		/* same as base above */
		long	num_elem;	/* number of elements */
		long	low_bnd;	/* lower bound set to 0 */
		long	high_bnd;	/* upper bound set to num_elem - 1 */

	} A_DSCR;

/* -------------------------------------------------------------------------- */
/* array of strings descriptor macros                                         */

#define $mkstr_array_l(ary, dat, dim, len) 	\
	VAR char	dat[dim * len];		\
	VAR A_DSCR	ary[1] = {		\
/* 	length   = */	len,			\
/* 	dtype    = */	DSC$K_DTYPE_T,		\
/* 	class    = */	DSC$K_CLASS_A,		\
/* 	base     = */	dat,			\
/* 	fill     = */	0,			\
/* 	flags    = */	0XD0,			\
/* 	dim_cnt  = */	1,			\
/* 	arsize   = */	dim * len,		\
/* 	addr0    = */	dat,			\
/* 	num_elem = */	dim,			\
/* 	low_bnd  = */	0,			\
/* 	high_bnd = */	dim - 1			\
	}

#define $mkstr_array_d(ary, dat, dim, len) 	\
	GDVAR char	dat[dim * len];		\
	GDVAR A_DSCR	ary[1] = {		\
/* 	length   = */	len,			\
/* 	dtype    = */	DSC$K_DTYPE_T,		\
/* 	class    = */	DSC$K_CLASS_A,		\
/* 	base     = */	dat,			\
/* 	fill     = */	0,			\
/* 	flags    = */	0XD0,			\
/* 	dim_cnt  = */	1,			\
/* 	arsize   = */	dim * len,		\
/* 	addr0    = */	dat,			\
/* 	num_elem = */	dim,			\
/* 	low_bnd  = */	0,			\
/* 	high_bnd = */	dim - 1			\
	}

#define $mkstr_array_r(ary, dat)		\
	GRVAR char	dat[];			\
	GRVAR A_DSCR	ary[];

#define mkstr_array(a, n, l) 			\
	{					\
	a = malloc (sizeof (A_DSCR));		\
	a->dtype    = DSC$K_DTYPE_T;		\
	a->class    = DSC$K_CLASS_A;		\
	a->flags    = 0XD0;			\
	a->dim_cnt  = 1;			\
	a->low_bnd  = 0;			\
	a->num_elem = n;			\
	a->high_bnd = n - 1;			\
	a->length   = l;			\
	a->arsize   = l * n;			\
	a->base     = calloc (n, l);		\
	a->addr0    = a->base;			\
	}

#define idx_elem(a, i)				\
	(((a)->base) + ((a)->length * (i)))

#define del_str_array(a)			\
	{					\
	cfree (a->base);			\
	free (a);				\
	}

/* -------------------------------------------------------------------------- */

	TYPE struct token {		/* size = 8 bytes */
		word	type;		/* token type */
		byte	len;		/* length of data in bytes */
		char	*data;		/* pointer to data */
		struct token	*next;	/* next node in list */
	} TOKEN, *TKNLST;

#define	NUL	((TKNLST) 0)		/* nul token pointer */

#define START	0			/* parse states and token types */
#define NORM	1
#define WILD	2
#define DONE	3

#define MATCH	1			/* match values */
#define NOMATCH	0

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* routine to create fixed length descriptor for character string             */

DSCR
*mkdscrl (dscr, str, len)

DSCR	*dscr;		/* pointer to data area to setup as a descriptor      */
char	*str;		/* pointer to start of character string               */
long	len;		/* length of the resultant character string           */
{
	dscr->length = len;
	dscr->type   = DSC$K_DTYPE_T;
	dscr->class  = DSC$K_CLASS_S;
	dscr->string = str;
	return dscr;	/* return pointer to descriptor                       */
}

long 
LIB_MATCH_DIR_SPEC (cand_str, pattern_str)
DSCR	*cand_str, *pattern_str;
{
	TKNLST	cand_tokens = NUL, pattern_tokens = NUL;
	long	status;

	get_tokens (&cand_tokens, cand_str);
	get_tokens (&pattern_tokens, pattern_str);

	status = match_tokens (cand_tokens, pattern_tokens);

	del_list (cand_tokens);
	del_list (pattern_tokens);

	return status;
}

/* -------------------------------------------------------------------------- */
static long 
match_names (cnd_lst, ptn_lst)
TKNLST	cnd_lst, ptn_lst;
{
	VAR DSCR dsc[1];
	globalvalue STR$_MATCH;

	return (STR$_MATCH == STR$MATCH_WILD (
		mkdscrl (&dsc[1], cnd_lst->data, cnd_lst->len), 
		mkdscrl (&dsc[0], ptn_lst->data, ptn_lst->len)));
}

/* -------------------------------------------------------------------------- */
static long 
match_tokens (cnd_lst, ptn_lst)
TKNLST	cnd_lst, ptn_lst;
{
	if (ptn_lst == NUL) {
	    if (cnd_lst == NUL) 
		return MATCH;
	    else 
		return NOMATCH;
	} else {
	    if (cnd_lst == NUL) return NOMATCH;
	    if (ptn_lst->type == WILD) {
		if (ptn_lst->next == NUL) return MATCH;
		while (cnd_lst != NUL && ! match_names (cnd_lst, ptn_lst->next))
		    cnd_lst = cnd_lst->next;
		if (cnd_lst == NUL) return NOMATCH;
		return match_tokens (cnd_lst->next, ptn_lst->next->next);
	    } else {
		if (! match_names (cnd_lst, ptn_lst)) return NOMATCH;
		return match_tokens (cnd_lst->next, ptn_lst->next);
	    };
	};
}

/* -------------------------------------------------------------------------- */
static TKNLST *
make_token (type, len, data)
word	type;
byte	len;
char	*data;
{
	register TKNLST ptr = malloc (sizeof (TOKEN));

	ptr->type = type;
	ptr->len = len;
	ptr->data = data;
	ptr->next = NUL;
	return ptr;
}

/* -------------------------------------------------------------------------- */
static 
del_list (list)
TKNLST	list;
{
	if (list == NUL) return;
	del_list (list->next);
	free (list);
}

/* -------------------------------------------------------------------------- */
static 
get_tokens (list_ptr, spec)
TKNLST	*list_ptr;
DSCR	*spec;
{
	register char cur_chr;
	register long
		 idx = 1
		,rpt_ctr = 1
		,cur_state
		,nxt_state = START
	;

	while ((cur_state = nxt_state) != DONE) {
	    cur_chr = spec->string [idx++];
	    if (cur_chr == '>' || cur_chr == ']') 
		nxt_state = DONE;
	    else if (cur_chr == '.') 
		nxt_state = WILD;
	    else
		nxt_state = NORM;
	    if (cur_state == nxt_state) rpt_ctr++;
	    if (cur_state != START && cur_state != nxt_state) {
		if (! (cur_state == WILD && rpt_ctr == 1)) {
		    *list_ptr = make_token (cur_state, rpt_ctr,
			    &spec->string[idx - rpt_ctr - 1]);
		    list_ptr = &(*list_ptr)->next;
		};
		rpt_ctr = 1;
	    };
	};
}

