/*	================================ MODOBJ ================================

	Modifies linker object files.  In particular,  GSD$C_SYM  and  GSD$C_EPM
	subrecords that define user-specified symbols are converted to GSD$C_PRO
	subrecords.  Similarly, GSD$C_SYMW and GSD$C_EPMW  subrecords  are  con-
	verted to GSD$C_PROW subrecords, and GSD$C_LSY and GSD$C_LEPM subrecords
	are converted to GSD$C_LPRO subrecords.

	Greg Janee, 09-OCT-1987

	------------------------------------------------------------------------

	1.  Background

	There are three types of Global Symbol  Directory  (GSD)  subrecords  in
	object  files  that  define  global  symbols.   They  differ in how much
	information is associated with the symbol being defined:   the  shortest
	assigns  only  a  value  to  the  symbol,  while the longest in addition
	declares that the symbol is the name  of  a  procedure,  and  lists  the
	number  of  arguments the procedure accepts and the calling mechanism of
	each argument.  Which subrecord type is used  depends  on  the  compiler
	that is producing the object file and the symbol's purpose.

	There  are at least two cases in which the linker needs more information
	about a symbol than just its value, but can't necessarily get  it.   The
	first case is the resolution of a .MASK directive in a MACRO module.  If
	the target entry point is declared with a .ENTRY  directive,  the  entry
	point's  mask  is  placed  in  the  symbol definition and the linker can
	resolve the directive.  If the entry point is declared  with  a  double-
	colon operator, though, as in

		entry_point::
			.WORD	mask

	the mask is not included in the symbol  definition  and  the  resolution
	fails.

	The  second case involves argument-passing mechanisms.  The FORTRAN com-
	piler uses the longest type of GSD subrecord for subroutine and function
	declarations,  and  hence  argument  information  is  present in FORTRAN
	object modules.  The linker uses this information to  determine  whether
	arguments  that  are  FORTRAN  character  constants  should be passed by
	reference (e.g. when the subroutine expects a Hollerith constant) or  by
	descriptor  (when  it  expects  a  character variable).  However, if the
	subroutine is written in MACRO, no argument  information  is  associated
	with the subroutine name and FORTRAN character constants are then passed
	by descriptor, the default.  The only recourse for the MACRO  subroutine
	that  is  to accept both FORTRAN Hollerith and character constants is to
	place %REF() operators around each character constant  argument  in  the
	calling routines, a solution that is not always feasible.

	2.  Function

	The MODOBJ  program  converts  symbol-defining  GSD  subrecords  to  the
	longest  subrecord  type,  type  GSD$C_PRO (or GSD$C_PROW or GSD$C_LPRO,
	depending on the source subrecord type).  The extra  symbol  information
	is  supplied  by  the user in the form of a datafile.  Both of the above
	problems are thereby circumvented.

	The companion file MODOBJ.CLD defines the MODOBJ command, which  accepts
	a  /DATAFILE=file  qualifier  identifying  the datafile to be used and a
	single parameter identifying an  object  file  containing  one  or  more
	object modules.  The default filetype for the datafile is .DAT; that for
	the object file is .OBJ.  MODOBJ processes the object file and  produces
	a higher version of the same file.  This new object file is functionally
	equivalent to the source object file,  except  that  any  GSD  subrecord
	defining one of the symbols specified in the datafile has been converted
	to a Procedure With Formal Argument Definition  (GSD$C_PRO,  GSD$C_PROW,
	or  GSD$C_LPRO)  subrecord.   The  information  required  to fill in the
	longer subrecord is taken  primarily  from  the  source  subrecord,  and
	secondarily from the datafile.

	If the /LOG qualifier is specified MODOBJ  displays  the  names  of  the
	symbols whose defining subrecords are modified.  The default is /NOLOG.

	3.  Datafile Format

	The datafile is a  sequential  file  with  variable-length  records  and
	carriage-return  carriage control.  The file contains zero or more data-
	file statements, whose syntax is formally defined below.   A  hyphen  at
	the  end  of a record is a continuation character and exclamation points
	are comment delimiters, as in DCL.  The statements are case-insensitive.
	Blank lines are ignored.

	<datafile_statement> ::= <procedure_name> ( <argument_specification> )
		| <procedure_name> ( )
		| <procedure_name> ( <argument_specification> ) MASK =
			<entry_mask_specification>
		| <procedure_name> ( ) MASK = <entry_mask_specification>

	<procedure_name> ::= <basic_symbol_character>
		| <basic_symbol_character><symbol_string>

	<symbol_string> ::= <symbol_character>
		| <symbol_character><symbol_string>

	<symbol_character> ::= <basic_symbol_character> | <digit>

	<basic_symbol_character> ::= <letter> | $ | _

	<letter> ::= A | B | C | D | E | F | G | H | I | J | K | L | M | N | O
		| P | Q | R | S | T | U | V | W | X | Y | Z

	<digit> ::= 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9

	<argument_specification> ::= <argument_list>
		| <argument_list> [, <argument_list> ]
		| [ <argument_list> ]

	<argument_list> ::= <argument_identifier>
		| <argument_identifier>, <argument_list>

	<argument_identifier> ::= R | D | V | U

	<entry_mask_specification> ::= ^M<> | ^M< <item_list> >

	<item_list> ::= <mask_item> | <mask_item>, <item_list>

	<mask_item> ::= R2 | R3 | R4 | R5 | R6 | R7 | R8 | R9 | R10 | R11 | IV
		| DV

	The procedure name identifies a symbol whose defining subrecord is to be
	converted;  the  argument specification and entry mask specification are
	the additional information needed to perform  the  conversion.   In  the
	argument  specification  each  letter  represents  both  an argument the
	procedure accepts and  the  calling  mechanism  of  the  argument.   The
	letters  R,  D,  V, and U indicate the reference, descriptor, value, and
	unknown mechanisms, respectively.  Arguments within the square  brackets
	are  optional.   If  there  is  no argument specification, the procedure
	accepts no arguments.  The entry mask specification indicates the  entry
	mask  of  the procedure, specified in standard MACRO format.  Specifying
	no entry mask is equivalent to specifying ^M<> for the mask.

	4.  Maintenance

	MODOBJ should be linked /NOTRACEBACK with SYNTAX.MAR and MESSAGES.MSG.

	======================================================================*/

/*============================================================================*/

#	include climsgdef
#	include ctype
#	include descrip
#	include fab
#	include libdef
#	include rab
#	include rmsdef
#	include ssdef
#	include stdio
#	include stsdef

/*----------------------------------------------------------------------------*/

/*	The  following  symbols  come  from  module $OBJDEF in SYS$LIBRARY:STAR-
	LET.MLB.  We include it here since it has no equivalent in  SYS$LIBRARY:
	VAXCDEF.TLB.   Note  that  the MODOBJ code makes assumptions about these
	symbols' values.
*/
#	define obj$c_hdr        0
#	define obj$c_hdr_mhd    0
#	define obj$c_hdr_lnm    1
#	define obj$c_hdr_src    2
#	define obj$c_hdr_ttl    3
#	define obj$c_hdr_cpr    4
#	define obj$c_hdr_mtc    5
#	define obj$c_hdr_gtx    6
#	define obj$c_gsd        1
#	define obj$c_gsd_psc    0
#	define obj$c_gsd_sym    1
#	define obj$c_gsd_epm    2
#	define obj$c_gsd_pro    3
#	define obj$c_gsd_symw   4
#	define obj$c_gsd_epmw   5
#	define obj$c_gsd_prow   6
#	define obj$c_gsd_idc    7
#	define obj$c_gsd_env    8
#	define obj$c_gsd_lsy    9
#	define obj$c_gsd_lepm   10
#	define obj$c_gsd_lpro   11
#	define obj$c_gsd_spsc   12
#	define obj$c_tir        2
#	define obj$c_eom        3
#	define obj$c_dbg        4
#	define obj$c_tbt        5
#	define obj$c_lnk        6
#	define obj$c_eomw       7
#	define obj$c_maxrectyp  7
#	define obj$k_subtyp     1
#	define obj$c_subtyp     1
#	define obj$c_maxrecsiz  2048
#	define obj$c_strlvl     0
#	define obj$c_symsiz     31
#	define obj$c_storeplim  -1
#	define obj$c_pscalilim  9
#	define obj$s_objrecdef  5
#	define obj$b_rectyp     0
#	define obj$b_subtyp     1
#	define obj$b_mhd_strlv  2
#	define obj$w_mhd_recsz  3
#	define obj$t_mhd_name   5
#	define mhd$c_mhd        0
#	define mhd$c_lnm        1
#	define mhd$c_src        2
#	define mhd$c_ttl        3
#	define mhd$c_cpr        4
#	define mhd$c_mtc        5
#	define mhd$c_gtx        6
#	define mhd$c_maxhdrtyp  6
#	define mhd$s_mhdef      37
#	define mhd$b_rectyp     0
#	define mhd$b_hdrtyp     1
#	define mhd$b_strlvl     2
#	define mhd$w_recsiz     3
#	define mhd$b_namlng     5
#	define mhd$s_name       31
#	define mhd$t_name       6
#	define eom$c_success    0
#	define eom$c_warning    1
#	define eom$c_error      2
#	define eom$c_abort      3
#	define eom$k_eommin     2
#	define eom$c_eommin     2
#	define eom$k_eommx1     7
#	define eom$c_eommx1     7
#	define eom$k_eommax     8
#	define eom$c_eommax     8
#	define eom$m_wktfr      1
#	define eom$s_eomdef     8
#	define eom$b_rectyp     0
#	define eom$b_comcod     1
#	define eom$b_psindx     2
#	define eom$l_tfradr     3
#	define eom$b_tfrflg     7
#	define eom$v_wktfr      0
#	define eomw$k_eommin    2
#	define eomw$c_eommin    2
#	define eomw$k_eommx1    8
#	define eomw$c_eommx1    8
#	define eomw$k_eommax    9
#	define eomw$c_eommax    9
#	define eomw$m_wktfr     1
#	define eomw$s_eomwdef   9
#	define eomw$b_rectyp    0
#	define eomw$b_comcod    1
#	define eomw$w_psindx    2
#	define eomw$l_tfradr    4
#	define eomw$b_tfrflg    8
#	define eomw$v_wktfr     0
#	define lnk$c_olb        0
#	define lnk$c_shr        1
#	define lnk$c_oli        2
#	define lnk$c_obj        3
#	define lnk$c_sha        4
#	define lnk$c_maxrectyp  4
#	define lnk$m_selser     1
#	define lnk$m_libsrch    2
#	define lnk$s_lnkdef     6
#	define lnk$b_rectyp     0
#	define lnk$b_lnktyp     1
#	define lnk$w_flags      2
#	define lnk$v_selser     0
#	define lnk$v_libsrch    1
#	define lnk$w_namlng     4
#	define lnk$t_name       6
#	define gsd$k_entries    1
#	define gsd$c_entries    1
#	define gsd$c_psc        0
#	define gsd$c_sym        1
#	define gsd$c_epm        2
#	define gsd$c_pro        3
#	define gsd$c_symw       4
#	define gsd$c_epmw       5
#	define gsd$c_prow       6
#	define gsd$c_idc        7
#	define gsd$c_env        8
#	define gsd$c_lsy        9
#	define gsd$c_lepm       10
#	define gsd$c_lpro       11
#	define gsd$c_spsc       12
#	define gsd$c_maxrectyp  12
#	define gsd$s_gsdef      2
#	define gsd$b_rectyp     0
#	define gsd$b_gsdtyp     1
#	define gps$m_pic        1
#	define gps$m_lib        2
#	define gps$m_ovr        4
#	define gps$m_rel        8
#	define gps$m_gbl        16
#	define gps$m_shr        32
#	define gps$m_exe        64
#	define gps$m_rd         128
#	define gps$m_wrt        256
#	define gps$m_vec        512
#	define gps$k_name       9
#	define gps$c_name       9
#	define gps$s_gpsdef     40
#	define gps$b_gsdtyp     0
#	define gps$t_start      0
#	define gps$b_align      1
#	define gps$w_flags      2
#	define gps$v_pic        0
#	define gps$v_lib        1
#	define gps$v_ovr        2
#	define gps$v_rel        3
#	define gps$v_gbl        4
#	define gps$v_shr        5
#	define gps$v_exe        6
#	define gps$v_rd         7
#	define gps$v_wrt        8
#	define gps$v_vec        9
#	define gps$l_alloc      4
#	define gps$b_namlng     8
#	define gps$s_name       31
#	define gps$t_name       9
#	define sgps$m_pic       1
#	define sgps$m_lib       2
#	define sgps$m_ovr       4
#	define sgps$m_rel       8
#	define sgps$m_gbl       16
#	define sgps$m_shr       32
#	define sgps$m_exe       64
#	define sgps$m_rd        128
#	define sgps$m_wrt       256
#	define sgps$m_vec       512
#	define sgps$k_name      13
#	define sgps$c_name      13
#	define sgps$s_sgpsdef   44
#	define sgps$b_gsdtyp    0
#	define sgps$t_start     0
#	define sgps$b_align     1
#	define sgps$w_flags     2
#	define sgps$v_pic       0
#	define sgps$v_lib       1
#	define sgps$v_ovr       2
#	define sgps$v_rel       3
#	define sgps$v_gbl       4
#	define sgps$v_shr       5
#	define sgps$v_exe       6
#	define sgps$v_rd        7
#	define sgps$v_wrt       8
#	define sgps$v_vec       9
#	define sgps$l_alloc     4
#	define sgps$l_base      8
#	define sgps$b_namlng    12
#	define sgps$s_name      31
#	define sgps$t_name      13
#	define gsy$m_weak       1
#	define gsy$m_def        2
#	define gsy$m_uni        4
#	define gsy$m_rel        8
#	define gsy$s_gsydef     4
#	define gsy$b_gsdtyp     0
#	define gsy$t_start      0
#	define gsy$b_datyp      1
#	define gsy$w_flags      2
#	define gsy$v_weak       0
#	define gsy$v_def        1
#	define gsy$v_uni        2
#	define gsy$v_rel        3
#	define srf$k_name       5
#	define srf$c_name       5
#	define srf$s_srfdef     36
#	define srf$b_gsdtyp     0
#	define srf$t_start      0
#	define srf$b_datyp      1
#	define srf$w_flags      2
#	define srf$b_namlng     4
#	define srf$s_name       31
#	define srf$t_name       5
#	define sdf$k_name       10
#	define sdf$c_name       10
#	define sdf$s_sdfdef     41
#	define sdf$b_gsdtyp     0
#	define sdf$t_start      0
#	define sdf$b_datyp      1
#	define sdf$w_flags      2
#	define sdf$b_psindx     4
#	define sdf$l_value      5
#	define sdf$b_namlng     9
#	define sdf$s_name       31
#	define sdf$t_name       10
#	define epm$k_name       12
#	define epm$c_name       12
#	define epm$s_epmdef     43
#	define epm$b_gsdtyp     0
#	define epm$t_start      0
#	define epm$b_datyp      1
#	define epm$w_flags      2
#	define epm$b_psindx     4
#	define epm$l_addrs      5
#	define epm$w_mask       9
#	define epm$b_namlng     11
#	define epm$s_name       31
#	define epm$t_name       12
#	define pro$k_name       12
#	define pro$c_name       12
#	define pro$s_prodef     43
#	define pro$b_gsdtyp     0
#	define pro$t_start      0
#	define pro$b_datyp      1
#	define pro$w_flags      2
#	define pro$b_psindx     4
#	define pro$l_addrs      5
#	define pro$w_mask       9
#	define pro$b_namlng     11
#	define pro$s_name       31
#	define pro$t_name       12
#	define fml$k_size       2
#	define fml$c_size       2
#	define fml$s_fmldef     2
#	define fml$b_minargs    0
#	define fml$b_maxargs    1
#	define arg$c_unknown    0
#	define arg$c_value      1
#	define arg$c_ref        2
#	define arg$c_desc       3
#	define arg$k_size       2
#	define arg$c_size       2
#	define arg$s_argdef     2
#	define arg$b_valctl     0
#	define arg$s_passmech   2
#	define arg$v_passmech   0
#	define arg$b_bytecnt    1
#	define sdfw$k_name      11
#	define sdfw$c_name      11
#	define sdfw$s_sdfwdef   42
#	define sdfw$b_gsdtyp    0
#	define sdfw$t_start     0
#	define sdfw$b_datyp     1
#	define sdfw$w_flags     2
#	define sdfw$w_psindx    4
#	define sdfw$l_value     6
#	define sdfw$b_namlng    10
#	define sdfw$s_name      31
#	define sdfw$t_name      11
#	define epmw$k_name      13
#	define epmw$c_name      13
#	define epmw$s_epmwdef   44
#	define epmw$b_gsdtyp    0
#	define epmw$t_start     0
#	define epmw$b_datyp     1
#	define epmw$w_flags     2
#	define epmw$w_psindx    4
#	define epmw$l_addrs     6
#	define epmw$w_mask      10
#	define epmw$b_namlng    12
#	define epmw$s_name      31
#	define epmw$t_name      13
#	define prow$k_name      13
#	define prow$c_name      13
#	define prow$s_prowdef   44
#	define prow$b_gsdtyp    0
#	define prow$t_start     0
#	define prow$b_datyp     1
#	define prow$w_flags     2
#	define prow$w_psindx    4
#	define prow$l_addrs     6
#	define prow$w_mask      10
#	define prow$b_namlng    12
#	define prow$s_name      31
#	define prow$t_name      13
#	define idc$c_leq        0
#	define idc$c_equal      1
#	define idc$s_idcdef     4
#	define idc$b_gsdtyp     0
#	define idc$w_flags      1
#	define idc$v_binident   0
#	define idc$s_idmatch    2
#	define idc$v_idmatch    1
#	define idc$s_errsev     3
#	define idc$v_errsev     3
#	define idc$b_namlng     3
#	define idc$t_name       4
#	define env$m_def        1
#	define env$m_nested     2
#	define env$s_envdef     37
#	define env$b_gsdtyp     0
#	define env$w_flags      1
#	define env$v_def        0
#	define env$v_nested     1
#	define env$w_envindx    3
#	define env$b_namlng     5
#	define env$s_name       31
#	define env$t_name       6
#	define lsy$m_weak       1
#	define lsy$m_def        2
#	define lsy$m_uni        4
#	define lsy$m_rel        8
#	define lsy$s_lsydef     6
#	define lsy$b_gsdtyp     0
#	define lsy$t_start      0
#	define lsy$b_datyp      1
#	define lsy$w_flags      2
#	define lsy$v_weak       0
#	define lsy$v_def        1
#	define lsy$v_uni        2
#	define lsy$v_rel        3
#	define lsy$w_envindx    4
#	define lsrf$k_name      7
#	define lsrf$c_name      7
#	define lsrf$s_lsrfdef   38
#	define lsrf$b_gsdtyp    0
#	define lsrf$t_start     0
#	define lsrf$b_datyp     1
#	define lsrf$w_flags     2
#	define lsrf$w_envindx   4
#	define lsrf$b_namlng    6
#	define lsrf$s_name      31
#	define lsrf$t_name      7
#	define lsdf$k_name      13
#	define lsdf$c_name      13
#	define lsdf$s_lsdfdef   44
#	define lsdf$b_gsdtyp    0
#	define lsdf$t_start     0
#	define lsdf$b_datyp     1
#	define lsdf$w_flags     2
#	define lsdf$w_envindx   4
#	define lsdf$w_psindx    6
#	define lsdf$l_value     8
#	define lsdf$b_namlng    12
#	define lsdf$s_name      31
#	define lsdf$t_name      13
#	define lepm$k_name      15
#	define lepm$c_name      15
#	define lepm$s_lepmdef   46
#	define lepm$b_gsdtyp    0
#	define lepm$t_start     0
#	define lepm$b_datyp     1
#	define lepm$w_flags     2
#	define lepm$w_envindx   4
#	define lepm$w_psindx    6
#	define lepm$l_addrs     8
#	define lepm$w_mask      12
#	define lepm$b_namlng    14
#	define lepm$s_name      31
#	define lepm$t_name      15
#	define lpro$k_name      15
#	define lpro$c_name      15
#	define lpro$s_lprodef   46
#	define lpro$b_gsdtyp    0
#	define lpro$t_start     0
#	define lpro$b_datyp     1
#	define lpro$w_flags     2
#	define lpro$w_envindx   4
#	define lpro$w_psindx    6
#	define lpro$l_addrs     8
#	define lpro$w_mask      12
#	define lpro$b_namlng    14
#	define lpro$s_name      31
#	define lpro$t_name      15
#	define tir$c_sta_gbl    0
#	define tir$c_sta_sb     1
#	define tir$c_sta_sw     2
#	define tir$c_sta_lw     3
#	define tir$c_sta_pb     4
#	define tir$c_sta_pw     5
#	define tir$c_sta_pl     6
#	define tir$c_sta_ub     7
#	define tir$c_sta_uw     8
#	define tir$c_sta_bfi    9
#	define tir$c_sta_wfi    10
#	define tir$c_sta_lfi    11
#	define tir$c_sta_epm    12
#	define tir$c_sta_ckarg  13
#	define tir$c_sta_wpb    14
#	define tir$c_sta_wpw    15
#	define tir$c_sta_wpl    16
#	define tir$c_sta_lsy    17
#	define tir$c_sta_lit    18
#	define tir$c_sta_lepm   19
#	define tir$c_maxstacod  19
#	define tir$c_minstocod  20
#	define tir$c_sto_sb     20
#	define tir$c_sto_sw     21
#	define tir$c_sto_l      22
#	define tir$c_sto_bd     23
#	define tir$c_sto_wd     24
#	define tir$c_sto_ld     25
#	define tir$c_sto_li     26
#	define tir$c_sto_pidr   27
#	define tir$c_sto_picr   28
#	define tir$c_sto_rsb    29
#	define tir$c_sto_rsw    30
#	define tir$c_sto_rl     31
#	define tir$c_sto_vps    32
#	define tir$c_sto_usb    33
#	define tir$c_sto_usw    34
#	define tir$c_sto_rub    35
#	define tir$c_sto_ruw    36
#	define tir$c_sto_b      37
#	define tir$c_sto_w      38
#	define tir$c_sto_rb     39
#	define tir$c_sto_rw     40
#	define tir$c_sto_rivb   41
#	define tir$c_sto_pirr   42
#	define tir$c_maxstocod  42
#	define tir$c_minoprcod  50
#	define tir$c_opr_nop    50
#	define tir$c_opr_add    51
#	define tir$c_opr_sub    52
#	define tir$c_opr_mul    53
#	define tir$c_opr_div    54
#	define tir$c_opr_and    55
#	define tir$c_opr_ior    56
#	define tir$c_opr_eor    57
#	define tir$c_opr_neg    58
#	define tir$c_opr_com    59
#	define tir$c_opr_insv   60
#	define tir$c_opr_ash    61
#	define tir$c_opr_ush    62
#	define tir$c_opr_rot    63
#	define tir$c_opr_sel    64
#	define tir$c_opr_redef  65
#	define tir$c_opr_dflit  66
#	define tir$c_maxoprcod  66
#	define tir$c_minctlcod  80
#	define tir$c_ctl_setrb  80
#	define tir$c_ctl_augrb  81
#	define tir$c_ctl_dfloc  82
#	define tir$c_ctl_stloc  83
#	define tir$c_ctl_stkdl  84
#	define tir$c_maxctlcod  84
#	define tir$s_tirdef     1
#	define tir$b_rectyp     0

/*----------------------------------------------------------------------------*/

	typedef unsigned char ubyte;
	typedef long unsigned int ulong;
	typedef short unsigned int uword;

	typedef char sbyte;
	typedef long int slong;
	typedef short int sword;

#	define MAX(x,y) ( ((x) < (y)) ? (y) : (x) )

/*	The  SYMBOL_INFORMATION_BLOCK structure is used to store the contents of
	the datafile.  Each block contains a procedure name, stored as  a  zero-
	terminated  string; the argument specification for the procedure, stored
	as a zero-terminated string of argument identifiers; an entry  mask  for
	the  procedure,  stored  as  an unsigned word; and the minimum number of
	arguments the procedure accepts, stored as an unsigned byte.
*/
	typedef struct symbol_information_block symblk;
	struct symbol_information_block {

		symblk	*next;
		char	name[32];
		char	args[256];
		uword	mask;
		ubyte	min_args;

	};

/*	The global variable SYMBOL_LISTHEAD heads a linked list of symbol infor-
	mation  blocks.  There is one block per symbol in the datafile.  LOGGING
	is a boolean that reflects the presence of the  /LOG  qualifier  on  the
	command line.
*/
	symblk	*SYMBOL_LISTHEAD;
	ubyte	LOGGING;

/*	The  following  structures  define  the  argument  block used by routine
	LIB$TPARSE.  The first structure is for referencing it when it  is  used
	as an argument list; the second structure is used to access it directly.
*/
#	define tpa$k_count0     8
#	define tpa$m_blanks     1
#	define tpa$m_abbrev     2
#	define tpa$m_abbrfm     4
#	define tpa$m_ambig      65536
#	define tpa$k_length0    36
#	define tpa$c_length0    36
#	define tpa$s_tpadef     36
#	define tpa$v_blanks     0
#	define tpa$v_abbrev     1
#	define tpa$v_abbrfm     2
#	define tpa$v_ambig      16

	typedef struct tparse_arg_block_as_arg_list tpaarg;
	struct tparse_arg_block_as_arg_list {

		ulong	tpa$l_options;
		ulong	tpa$l_stringcnt;
		char	*tpa$l_stringptr;
		ulong	tpa$l_tokencnt;
		char	*tpa$l_tokenptr;
		char	tpa$b_char;
		ulong	: 24;
		ulong	tpa$l_number;
		ulong	tpa$l_param;

	};

	typedef struct tparse_argument_block tpablk;
	struct tparse_argument_block {

		ulong	tpa$l_count;
		ulong	tpa$l_options;
		ulong	tpa$l_stringcnt;
		char	*tpa$l_stringptr;
		ulong	tpa$l_tokencnt;
		char	*tpa$l_tokenptr;
		char	tpa$b_char;
		ulong	: 24;
		ulong	tpa$l_number;
		ulong	tpa$l_param;

	};

/*	Our  referenced  global  symbols  include  various  condition values and
	TPA_STA_TAB and  TPA_KEY_TAB,  used  to  locate  the  LIB$TPARSE  syntax
	tables.
*/
	globalvalue MODOBJ$_DE0;
	globalvalue MODOBJ$_DE1;
	globalvalue MODOBJ$_DE2;
	globalvalue MODOBJ$_EOF;
	globalvalue MODOBJ$_IOM;
	globalvalue MODOBJ$_IPN;
	globalvalue MODOBJ$_MOD;
	globalvalue MODOBJ$_OFE;
	globalvalue MODOBJ$_STL;
	globalvalue MODOBJ$_TMA;
	globalvalue MODOBJ$_UST;
	globalvalue MODOBJ$_UTO;

	globalvalue TPA_STA_TAB;
	globalvalue TPA_KEY_TAB;

/*	The  maximum  size of a GSD record containing a single GSD$C_PRO, GSD$C_
	PROW, or GSD$C_LPRO subrecord...
*/
#	define MAX_GSD_PRO_LEN ( 1 + lpro$s_lprodef + 2 + (255*2) )

/*============================================================================*/

/*============================================================================*/

	ulong	main ( )
{
	char	filename[256];
	$DESCRIPTOR( filename_desc, filename );
	struct	FAB inp_fab;
	struct	RAB inp_rab;
	ulong	length;
	uword	*maxptr;
	struct	FAB out_fab;
	struct	RAB out_rab;
	$DESCRIPTOR( q_datafile, "DATAFILE" );
	$DESCRIPTOR( q_log, "LOG" );
	$DESCRIPTOR( q_object, "OBJECT" );
	ubyte	record[32768];
	ulong	status;

	void	cli$get_value();
	ulong	cli$present();
	void	lib$signal();
	void	load_datafile();
	void	ots$move5();
	void	remove_first_subrecord();
	ulong	strlen();
	ulong	sys$close();
	ulong	sys$connect();
	ulong	sys$create();
	ulong	sys$get();
	ulong	sys$open();
	void	write_first_subrecord();
	void	write_record();

/*----------------------------------------------------------------------------*/

/*	The first  thing to do  is get the datafile's  filename from the command
	line and load  the file's contents into  our symbol list.  Note that the
	/DATAFILE qualifier is required.
*/
	length = 0; cli$get_value( &q_datafile, &filename_desc, &length );
	filename[length] = '\0';

	SYMBOL_LISTHEAD = NULL;
	load_datafile( filename );

	LOGGING = FALSE;
	if ( cli$present( &q_log ) == CLI$_PRESENT ) LOGGING = TRUE;

	length = 0; cli$get_value( &q_object, &filename_desc, &length );
	filename[length] = '\0';

/*----------------------------------------------------------------------------*/

/*	Initialize two pairs of access blocks to be used in reading in the  exi-
	sting object file and writing the new.
*/
	ots$move5( 0, 0, 0, FAB$C_BLN, &inp_fab );

	inp_fab.fab$b_bid = FAB$C_BID;
	inp_fab.fab$b_bln = FAB$C_BLN;
	inp_fab.fab$l_dna = ".OBJ";
	inp_fab.fab$b_dns = 4;
	inp_fab.fab$b_fac = FAB$M_GET;
	inp_fab.fab$l_fna = filename;
	inp_fab.fab$b_fns = strlen( filename );
	inp_fab.fab$l_fop = FAB$M_SQO;
	inp_fab.fab$b_shr = FAB$M_SHRGET;

	ots$move5( 0, 0, 0, RAB$C_BLN, &inp_rab );

	inp_rab.rab$b_bid = RAB$C_BID;
	inp_rab.rab$b_bln = RAB$C_BLN;
	inp_rab.rab$l_fab = &inp_fab;
	inp_rab.rab$b_rac = RAB$C_SEQ;
	inp_rab.rab$l_ubf = record;
	inp_rab.rab$w_usz = sizeof( record );

	ots$move5( 0, 0, 0, FAB$C_BLN, &out_fab );

	out_fab.fab$b_bid = FAB$C_BID;
	out_fab.fab$b_bln = FAB$C_BLN;
	out_fab.fab$l_dna = ".OBJ";
	out_fab.fab$b_dns = 4;
	out_fab.fab$b_fac = FAB$M_PUT;
	out_fab.fab$l_fna = filename;
	out_fab.fab$b_fns = strlen( filename );
	out_fab.fab$l_fop = ( FAB$M_MXV | FAB$M_SQO | FAB$M_TEF );
	out_fab.fab$b_org = FAB$C_SEQ;
	out_fab.fab$b_rfm = FAB$C_VAR;
	out_fab.fab$b_shr = FAB$M_NIL;

	ots$move5( 0, 0, 0, RAB$C_BLN, &out_rab );

	out_rab.rab$b_bid = RAB$C_BID;
	out_rab.rab$b_bln = RAB$C_BLN;
	out_rab.rab$l_fab = &out_fab;
	out_rab.rab$b_rac = RAB$C_SEQ;

/*	Finally,  open the specified object file, and create a higher version of
	it.
*/
	status = sys$open( &inp_fab );

	if ( (status & STS$M_SUCCESS) == 0 ) {
		lib$signal( MODOBJ$_OFE, 0, status, inp_fab.fab$l_stv );
	}

	status = sys$connect( &inp_rab );

	if ( (status & STS$M_SUCCESS) == 0 ) {
		lib$signal( MODOBJ$_OFE, 0, status, inp_rab.rab$l_stv );
	}

	status = sys$create( &out_fab );

	if ( (status & STS$M_SUCCESS) == 0 ) {
		lib$signal( MODOBJ$_OFE, 0, status, out_fab.fab$l_stv );
	}

	status = sys$connect( &out_rab );

	if ( (status & STS$M_SUCCESS) == 0 ) {
		lib$signal( MODOBJ$_OFE, 0, status, out_rab.rab$l_stv );
	}

/*----------------------------------------------------------------------------*/

	while ( (status = sys$get( &inp_rab )) != RMS$_EOF ) {

		if ( (status & STS$M_SUCCESS) == 0 ) {
			lib$signal( MODOBJ$_OFE, 0, status, inp_rab.rab$l_stv );
		}

		length = inp_rab.rab$w_rsz;

		if      ( record[obj$b_rectyp] == obj$c_hdr &&
		          record[mhd$b_hdrtyp] == mhd$c_mhd    ) {

/*			This is  the Main Module  Header Record.  We  check that
			the structure  level is compatible  with the version  of
			this  program, and we make sure  that the maximum record
			size can handle a GSD record containing a single subrec-
			ord of type GSD$C_LPRO.
*/
			if ( record[mhd$b_strlvl] != obj$c_strlvl ) {
				lib$signal( MODOBJ$_OFE, 0, MODOBJ$_IOM );
			}

			maxptr  = &record[mhd$w_recsiz];
			*maxptr = MAX( *maxptr, MAX_GSD_PRO_LEN );

			write_record( record, length, &out_rab );

		}
		else if ( record[obj$b_rectyp] == obj$c_gsd ) {

/*			This is  a Global Symbol Directory Record.  We  break it
			apart  into its  constituent  subrecords and analyze and
			write out each subrecord as an independent GSD record.
*/
			while ( length > 1 ) {
				write_first_subrecord( record, &out_rab );
				remove_first_subrecord( record, &length );
			}

		}
		else {

/*			This is some other  type of record,  and hence we simply
			write it back out.
*/
			write_record( record, length, &out_rab );

		}

	}

/*----------------------------------------------------------------------------*/

	status = sys$close( &inp_fab );

	if ( (status & STS$M_SUCCESS) == 0 ) {
		lib$signal( MODOBJ$_OFE, 0, status, inp_fab.fab$l_stv );
	}

	status = sys$close( &out_fab );

	if ( (status & STS$M_SUCCESS) == 0 ) {
		lib$signal( MODOBJ$_OFE, 0, status, out_fab.fab$l_stv );
	}

	return ( SS$_NORMAL );
}
/*============================================================================*/

/*============================================================================*/

	void	addbyte ( byte, dstptr )

/*	Appends a single byte to a byte string.  The pointer specifying the byte
	string is incremented.
*/
	ubyte	byte;
	ubyte	*(*dstptr);
{
	ubyte	*srcptr;

	void	movstr();

/*----------------------------------------------------------------------------*/

	srcptr = &byte;
	movstr( 1, &srcptr, dstptr );

	return;
}
/*============================================================================*/

/*============================================================================*/

	symblk	*get_block ( name )

/*	Searches the symbol information block list for the block associated with
	the  specified procedure name.  If the block is found a pointer to it is
	returned; otherwise, the null pointer is returned.
*/
	char	name[];
{
	symblk	*ptr;

	slong	strcmp();

/*----------------------------------------------------------------------------*/

	ptr = SYMBOL_LISTHEAD;

	while ( ptr != NULL && strcmp( ptr->name, name ) != 0 ) {
		ptr = ptr->next;
	}

	return ( ptr );
}
/*============================================================================*/

/*============================================================================*/

	ulong	ignore_blanks ( arg )

/*	Turns LIB$TPARSE's interpretation of blanks off.  This routine is called
	by LIB$TPARSE when specified to do so by a transition.
*/
	tpaarg	arg;
{
/*----------------------------------------------------------------------------*/

	arg.tpa$l_options &= ~tpa$m_blanks;
	return ( SS$_NORMAL );
}
/*============================================================================*/

/*============================================================================*/

	ulong	interpret_blanks ( arg )

/*	Turns  LIB$TPARSE's interpretation of blanks on.  This routine is called
	by LIB$TPARSE when specified to do so by a transition.
*/
	tpaarg	arg;
{
/*----------------------------------------------------------------------------*/

	arg.tpa$l_options |= tpa$m_blanks;
	return ( SS$_NORMAL );
}
/*============================================================================*/

/*============================================================================*/

	ulong	length_gsd_subrecord ( subrec )

/*	Returns the length of  a subrecord belonging to a GSD record.  The units
	are bytes.
*/
	ubyte	subrec[];
{
	ulong	arg_cnt;
	uword	*flag_ptr;
	ulong	i;
	ulong	subreclen;

	void	lib$signal();

/*----------------------------------------------------------------------------*/

	switch ( subrec[0] ) {

	case gsd$c_psc:

/*		Program Section Definition Subrecord.  The subrecord consists of
		8 bytes followed by a counted ASCII string.
*/
		subreclen  = 9;
		subreclen += subrec[gps$b_namlng];

		break;

	case gsd$c_sym:

/*		Global Symbol  Specification Subrecord.  The subrecord has defi-
		nition and reference  forms, as  indicated by  the GSY$V_DEF bit
		in the GSY$W_FLAGS field.  In either case it consists of a fixed
		portion followed by a counted ASCII string.
*/
		subreclen = 4;
		flag_ptr  = &subrec[gsy$w_flags];

		if ( ((*flag_ptr) & gsy$m_def) != 0 ) {
			subreclen += 6;
			subreclen += subrec[sdf$b_namlng];
		}
		else {
			subreclen += 1;
			subreclen += subrec[srf$b_namlng];
		}

		break;

	case gsd$c_epm:

/*		Entry Point Symbol and Mask Definition Subrecord.  The subrecord
		consists of 11 bytes followed by a counted ASCII string.
*/
		subreclen  = 12;
		subreclen += subrec[epm$b_namlng];

		break;

	case gsd$c_pro:

/*		Procedure With Formal Argument Definition Subrecord.  The subre-
		cord consists of 11 bytes, a counted ASCII  string, 2 bytes, and
		a formal argument descriptor for each argument.  Each descriptor
		consists of 1 byte followed by a counted ASCII string.
*/
		subreclen  = 12;
		subreclen += subrec[pro$b_namlng];
		subreclen += 2;

		arg_cnt = subrec[(subreclen-2)+fml$b_maxargs];

		for ( i = 0; i < arg_cnt; ++i ) {

			subreclen += 2;
			subreclen += subrec[subreclen-1];

		}

		break;

	case gsd$c_symw:

/*		Symbol Definition With  Word Psect Subrecord.  This subrecord is
		identical to the definition form of  type GSD$C_SYM except  that
		it is 1 byte longer.
*/
		subreclen  = 11;
		subreclen += subrec[sdfw$b_namlng];

		break;

	case gsd$c_epmw:

/*		Entry Point Definition With  Word Psect Subrecord.  This subrec-
		ord is identical to type GSD$C_EPM except that it is 1 byte lon-
		ger.
*/
		subreclen  = 13;
		subreclen += subrec[epmw$b_namlng];

		break;

	case gsd$c_prow:

/*		Procedure Definition With  Word Psect Subrecord.  This subrecord
		is  identical to type GSD$C_PRO  except that  its initial  fixed
		portion is 1 byte longer.
*/
		subreclen  = 13;
		subreclen += subrec[prow$b_namlng];
		subreclen += 2;

		arg_cnt = subrec[(subreclen-2)+fml$b_maxargs];

		for ( i = 0; i < arg_cnt; ++i ) {

			subreclen += 2;
			subreclen += subrec[subreclen-1];

		}

		break;

	case gsd$c_idc:

/*		Entity Ident  Consistency Check  Subrecord.  The  subrecord con-
		sists of 3 bytes followed by 3 counted ASCII strings.
*/
		subreclen  = 3;

		subreclen += 1;
		subreclen += subrec[subreclen-1];

		subreclen += 1;
		subreclen += subrec[subreclen-1];

		subreclen += 1;
		subreclen += subrec[subreclen-1];

		break;

	case gsd$c_env:

/*		Environment  Definition/Reference Subrecord.  The subrecord con-
		sists of 5 bytes followed by a counted ASCII string.
*/
		subreclen  = 6;
		subreclen += subrec[env$b_namlng];

		break;

	case gsd$c_lsy:

/*		Module-Local Symbol Definition/Reference Subrecord.  The defini-
		tion form of this subrecord is identical to  type GSD$C_SYMW and
		the  reference form  is identical to  the reference form of type
		GSD$C_SYM, except that in each case it is 2 bytes longer.
*/
		subreclen = 4;
		flag_ptr  = &subrec[lsy$w_flags];

		if ( ((*flag_ptr) & lsy$m_def) != 0 ) {
			subreclen += 9;
			subreclen += subrec[lsdf$b_namlng];
		}
		else {
			subreclen += 3;
			subreclen += subrec[lsrf$b_namlng];
		}

		break;

	case gsd$c_lepm:

/*		Module-Local Entry  Point Definition  Subrecord.  This subrecord
		is identical to type GSD$C_EPMW  except that it is 2 bytes long-
		er.
*/
		subreclen  = 15;
		subreclen += subrec[lepm$b_namlng];

		break;

	case gsd$c_lpro:

/*		Module-Local Procedure  Definition Subrecord.  This subrecord is
		identical to type GSD$C_PROW except that  its initial fixed por-
		tion is 2 bytes longer.
*/
		subreclen  = 15;
		subreclen += subrec[lpro$b_namlng];
		subreclen += 2;

		arg_cnt = subrec[(subreclen-2)+fml$b_maxargs];

		for ( i = 0; i < arg_cnt; ++i ) {

			subreclen += 2;
			subreclen += subrec[subreclen-1];

		}

		break;

	case gsd$c_spsc:

/*		Program Section Definition  in a Shareable Image.  The subrecord
		is identical to type GSD$C_PSC except that it is 4 bytes longer.
		It cannot reside  in an object file, but we include it here any-
		way.
*/
		subreclen  = 13;
		subreclen += subrec[sgps$b_namlng];

		break;

	default:

		lib$signal( MODOBJ$_OFE, 0, MODOBJ$_UST );

	}

	return ( subreclen );
}
/*============================================================================*/

/*============================================================================*/

	void	load_datafile ( filename )

/*	Loads  the contents of the specified MODOBJ datafile into a list of sym-
	bol information blocks headed by the  global  variable  SYMBOL_LISTHEAD.
	SYMBOL_LISTHEAD is not cleared beforehand.
*/
	char	filename[];
{
	FILE	*datafile;
	ulong	i;
	char	line[32768];
	ulong	line_no;
	symblk	*newblk;
	tpablk	parse_block;
	ulong	status;

	slong	fclose();
	FILE	*fopen();
	void	lib$signal();
	ulong	lib$tparse();
	ubyte	*malloc();
	ulong	next_line();
	ulong	strlen();
	ulong	strspn();

/*----------------------------------------------------------------------------*/

	if ( (datafile = fopen( filename, "r", "dna=.DAT" )) == NULL ) {
		lib$signal( MODOBJ$_DE0, 0, MODOBJ$_UTO );
	}

	line_no = 1;
	while ( next_line( line, sizeof( line ), datafile, line_no ) ) {

		if ( strspn( line, " \t" ) != strlen( line ) ) {

/*		We've  got  a datafile statement (consisting of one or more rec-
		ords) that has been stripped of comments  and  is  not  entirely
		blank.  Before parsing it we convert it to uppercase.
*/
		i = 0;
		while ( (line[i] = _toupper( line[i] )) != '\0' ) ++i;

/*		Allocate a symbol information block and initialize its  contents
		and place it at the head of the symbol list.
*/
		newblk = malloc( sizeof( symblk ) );
		if ( newblk == NULL ) lib$signal( LIB$_INSVIRMEM );

		newblk->name[0]  = '\0';
		newblk->args[0]  = '\0';
		newblk->mask     = 0;
		newblk->min_args = 0;

		newblk->next     = SYMBOL_LISTHEAD;
		SYMBOL_LISTHEAD  = newblk;

/*		(Re)initialize  the  parse argument block and call LIB$TPARSE to
		parse the statement.  Our various action routines will move  the
		contents of the line into the  new  symbol block  via the global
		variable SYMBOL_LISTHEAD.
*/
		parse_block.tpa$l_count     = tpa$k_count0;
		parse_block.tpa$l_options   = 0;
		parse_block.tpa$l_stringcnt = strlen( line );
		parse_block.tpa$l_stringptr = line;

		status = lib$tparse( &parse_block, TPA_STA_TAB, TPA_KEY_TAB );

/*		If the parsing operation failed, we identify the syntax error to
		the user by printing the line number of the line, and  the posi-
		tion within the line, at which it occurred.
*/
		if ( (status & STS$M_SUCCESS) == 0 ) {

			i = parse_block.tpa$l_tokenptr - line + 1;
			lib$signal( MODOBJ$_DE2, 2, line_no, i, status );

		}

		}

		++line_no;

	}

	fclose( datafile );
	return;
}
/*============================================================================*/

/*============================================================================*/

	void	movstr ( length, srcptr, dstptr )

/*	Moves a source byte string to a destination string.  The pointers speci-
	fying the  strings are updated to point to the bytes immediately follow-
	ing the strings.
*/
	ulong	length;
	ubyte	*(*srcptr);
	ubyte	*(*dstptr);
{
	void	ots$move3();

/*----------------------------------------------------------------------------*/

	ots$move3( length, *srcptr, *dstptr );

	*srcptr += length;
	*dstptr += length;

	return;
}
/*============================================================================*/

/*============================================================================*/

	ulong	next_line ( line, length, file, line_no )

/*	Reads in the next statement from the specified file stream.  A statement
	generally corresponds to the next  sequence  of  characters  up  to  and
	including  the  next newline, except that exclamation points and hyphens
	are interpreted.  Specifically, if an exclamation point is found  within
	a  statement,  it  and any characters following it are truncated; if the
	last non-white-space character in a statement is a  hyphen,  the  hyphen
	is replaced by the next statement.  Newlines are truncated.  The LINE_NO
	argument is the number of the statement, for error signaling purposes.
*/
	char	line[];
	ulong	length;
	FILE	*file;
	ulong	line_no;
{
	char	*ptr;

	char	*fgets();
	void	lib$signal();
	char	*strchr();
	ulong	strlen();
	char	*strrchr();
	ulong	strspn();

/*----------------------------------------------------------------------------*/

	if ( fgets( line, length, file ) == NULL ) return ( 0 );
	else {

/*		FGETS  puts  the  terminating newline in the read buffer, so its
		absence indicates a buffer overflow.  We  shorten  the  line  by
		getting rid of the newline and disregarding any comment.
*/
		if ( (ptr = strchr( line, '\n' )) == NULL ) {
			lib$signal( MODOBJ$_DE1, 1, line_no, MODOBJ$_STL );
		}
		else {
			*ptr = '\0';
		}

		if ( (ptr = strchr( line, '!' )) != NULL ) *ptr = '\0';

/*		Continuation record processing.  If  there's  a  hyphen  in  the
		record, and if the only characters that follow it are blanks and
		tabs, we read in another record and overwrite the hyphen.
*/
		if ( (ptr = strrchr( line, '-' )) != NULL && strspn( ptr+1,
		     " \t" ) == strlen( ptr+1 ) ) {

		if ( next_line( ptr, length-(ptr-line), file, line_no ) == 0 ) {
			lib$signal( MODOBJ$_DE1, 1, line_no, MODOBJ$_EOF );
		}

		}

		return ( 1 );

	}
}
/*============================================================================*/

/*============================================================================*/

	void	remove_first_subrecord ( record, reclen )

/*	Given a GSD record containing one or  more subrecords, removes the first
	subrecord.  The record's length is updated accordingly.
*/
	ubyte	record[];
	ulong	*reclen;
{
	ulong	movelen;
	ulong	subreclen;

	ulong	length_gsd_subrecord();
	void	ots$move3();

/*----------------------------------------------------------------------------*/

	subreclen = length_gsd_subrecord( &record[1] );
	movelen   = (*reclen) - 1 - subreclen;

/*	Note that if we're  removing the record's only subrecord the move length
	will be zero and the following call will do nothing.
*/
	ots$move3( movelen, &record[1+subreclen], &record[1] );
	*reclen -= subreclen;

	return;
}
/*============================================================================*/

/*============================================================================*/

	ulong	store_argument_identifier ( arg )

/*	Stores an argument identifier in the symbol information block pointed to
	by SYMBOL_LISTHEAD.  This routine is  called  by  LIB$TPARSE,  once  per
	argument  per  procedure.  If the TPA$L_PARAM field is 0 the argument is
	required; if it is 1 the argument is optional.
*/
	tpaarg	arg;
{
	ulong	length;

	ulong	strlen();

/*----------------------------------------------------------------------------*/

	if ( arg.tpa$l_param == 0 ) ++SYMBOL_LISTHEAD->min_args;

/*	We check here if  we've exceeded our limit of 255 arguments, which coin-
	cides with the VAX's architectural limitation.
*/
	length = strlen( SYMBOL_LISTHEAD->args );

	if ( length == 255 ) return ( MODOBJ$_TMA );
	else {

		SYMBOL_LISTHEAD->args[length]   = arg.tpa$b_char;
		SYMBOL_LISTHEAD->args[length+1] = '\0';

		return ( SS$_NORMAL );

	}
}
/*============================================================================*/

/*============================================================================*/

	ulong	store_mask_item ( arg )

/*	Stores an entry mask item in the mask field of  the  symbol  information
	block  pointed  to  by  SYMBOL_LISTHEAD.   This  routine  is  called  by
	LIB$TPARSE, once per mask item per procedure.
*/
	tpaarg	arg;
{
/*----------------------------------------------------------------------------*/

	SYMBOL_LISTHEAD->mask |= arg.tpa$l_param;
	return ( SS$_NORMAL );
}
/*============================================================================*/

/*============================================================================*/

	ulong	store_procedure_name ( arg )

/*	Stores  a symbol name  in the symbol  information  block  pointed  to by
	SYMBOL_LISTHEAD.  This routine  is called by LIB$TPARSE when it encount-
	ers a symbol name at the beginning of a datafile line.
*/
	tpaarg	arg;
{
	ulong	length;

	void	ots$move3();

/*----------------------------------------------------------------------------*/

/*	Note that  we still need  to check that the  symbol is no longer than 31
	characters and that it does not start with a digit.
*/
	length = arg.tpa$l_tokencnt;

	if ( length > 31 || isdigit( *arg.tpa$l_tokenptr ) ) {

		return ( MODOBJ$_IPN );

	}
	else {

		ots$move3( length, arg.tpa$l_tokenptr, SYMBOL_LISTHEAD->name );
		SYMBOL_LISTHEAD->name[length] = '\0';

		return ( SS$_NORMAL );

	}
}
/*============================================================================*/

/*============================================================================*/

	char	*symbol_name ( subrec, name )

/*	Returns the  symbol name associated  with a specified GSD subrecord as a
	zero-terminated  string.  The  return value is  a pointer to  NAME.  The
	largest possible name size is 31 bytes, not including terminator.
*/
	ubyte	subrec[];
	char	name[];
{
	uword	*flag_ptr;
	ulong	i;
	ulong	length;

	void	lib$signal();
	void	ots$move3();

/*----------------------------------------------------------------------------*/

/*	All symbol names are stored as  counted ASCII strings.  For each subrec-
	ord type we  locate where the name proper is stored; the byte count will
	then be in the previous byte.
*/
	switch ( subrec[0] ) {

	case gsd$c_psc:

		i = gps$t_name;
		break;

	case gsd$c_sym:

		flag_ptr = &subrec[gsy$w_flags];

		if ( ((*flag_ptr) & gsy$m_def) != 0 ) {
			i = sdf$t_name;
		}
		else {
			i = srf$t_name;
		}

		break;

	case gsd$c_epm:

		i = epm$t_name;
		break;

	case gsd$c_pro:

		i = pro$t_name;
		break;

	case gsd$c_symw:

		i = sdfw$t_name;
		break;

	case gsd$c_epmw:

		i = epmw$t_name;
		break;

	case gsd$c_prow:

		i = prow$t_name;
		break;

	case gsd$c_idc:

		i = idc$t_name;
		break;

	case gsd$c_env:

		i = env$t_name;
		break;

	case gsd$c_lsy:

		flag_ptr = &subrec[lsy$w_flags];

		if ( ((*flag_ptr) & lsy$m_def) != 0 ) {
			i = lsdf$t_name;
		}
		else {
			i = lsrf$t_name;
		}

		break;

	case gsd$c_lepm:

		i = lepm$t_name;
		break;

	case gsd$c_lpro:

		i = lpro$t_name;
		break;

	case gsd$c_spsc:

		i = sgps$t_name;
		break;

	default:

		lib$signal( MODOBJ$_OFE, 0, MODOBJ$_UST );

	}

/*----------------------------------------------------------------------------*/

	length = subrec[i-1];

	ots$move3( length, &subrec[i], name );
	name[length] = '\0';

	return ( name );
}
/*============================================================================*/

/*============================================================================*/

	void	transfer_contents ( subrec, mask, newrec )

/*	Performs the following transfers of GSD subrecord data:  GSD$C_SYM (def-
	inition form) and GSD$C_EPM to GSD$C_PRO; GSD$C_SYMW and  GSD$C_EPMW  to
	GSD$C_PROW;  and  GSD$C_LSY  (definition  form) and GSD$C_LEPM to GSD$C_
	LPRO.  The resulting Procedure With Formal Argument Definition Subrecord
	is complete up to and including the NAME field.  If the source subrecord
	does not include entry mask information, the MASK argument is used.
*/
	ubyte	subrec[];
	uword	mask;
	ubyte	*(*newrec);
{
	ulong	movelen;
	ubyte	rectype;
	ubyte	*srcptr;
	ubyte	*tmpptr;

	void	addbyte();
	void	movstr();

/*----------------------------------------------------------------------------*/

/*	The six valid input subrecord types are identical with three exceptions:
	the  GSD$C_SYMW  and  GSD$C_EPMW  types have a word (as opposed to byte)
	PSINDX field; the GSD$C_EPM, GSD$C_EPMW, and  GSD$C_LEPM  types  have  a
	MASK field; and the GSD$C_LSY and GSD$C_LEPM types have an ENVINDX field
	as well as a word PSINDX field.
*/
	rectype = gsd$c_pro;
	movelen = 8;

	if      ( subrec[0] == gsd$c_symw || subrec[0] == gsd$c_epmw ) {
		rectype  = gsd$c_prow;
		movelen += 1;
	}
	else if ( subrec[0] == gsd$c_lsy  || subrec[0] == gsd$c_lepm ) {
		rectype  = gsd$c_lpro;
		movelen += 3;
	}

	addbyte( rectype, newrec );

	srcptr = &subrec[1];
	movstr( movelen, &srcptr, newrec );

	if ( subrec[0] == gsd$c_epm || subrec[0] == gsd$c_epmw || subrec[0] ==
	     gsd$c_lepm ) {
		movstr( 2, &srcptr, newrec );
	}
	else {
		tmpptr = &mask;
		movstr( 2, &tmpptr, newrec );
	}

/*	To finish off the  initial part  of the new subrecord we move the symbol
	name, which is stored as a counted ASCII string.
*/
	movstr( 1+(*srcptr), &srcptr, newrec );
	return;
}
/*============================================================================*/

/*============================================================================*/

	void	write_first_subrecord ( record, rab )

/*	Analyzes  and writes to the specified file the first subrecord belonging
	to a given GSD record.  If the subrecord is one of  the  shorter  symbol
	definition subrecords, and if the defined symbol is on the symbol infor-
	mation block list, the subrecord is converted to a  GSD$C_(L)PRO(W)-type
	subrecord before being written out.
*/
	ubyte	record[];
	struct	RAB *rab;
{
	ulong	arg_cnt;
	symblk	*block;
	ubyte	doit;
	uword	*flag_ptr;
	ulong	i;
	char	name[32];
	ubyte	newrec[MAX_GSD_PRO_LEN];
	ubyte	*recptr;

	void	addbyte();
	symblk	*get_block();
	ulong	length_gsd_subrecord();
	void	lib$signal();
	ulong	strlen();
	char	*symbol_name();
	void	transfer_contents();
	void	write_record();

/*----------------------------------------------------------------------------*/

/*	The first step towards determining whether we should create a new subre-
	cord or not is to see if the source  subrecord has one of the acceptable
	types.
*/
	doit = FALSE;

	if      ( record[gsd$b_gsdtyp] == gsd$c_sym  ||
	          record[gsd$b_gsdtyp] == gsd$c_symw ||
	          record[gsd$b_gsdtyp] == gsd$c_lsy     ) {

		flag_ptr = &record[1+gsy$w_flags];
		if ( ((*flag_ptr) & gsy$m_def) != 0 ) doit = TRUE;

	}
	else if ( record[gsd$b_gsdtyp] == gsd$c_epm  ||
	          record[gsd$b_gsdtyp] == gsd$c_epmw ||
	          record[gsd$b_gsdtyp] == gsd$c_lepm    ) {

		doit = TRUE;

	}

/*	Second step:  if the  subrecord is  acceptable, see if we have an infor-
	mation block for the subrecord's symbol.
*/
	if ( doit ) {

		block = get_block( symbol_name( &record[1], name ) );
		if ( block == NULL ) doit = FALSE;

	}

/*----------------------------------------------------------------------------*/

	if ( doit ) {

/*		We're  making  a  new  GSD subrecord.  First set the identifying
		byte to OBJ$C_GSD, then transfer the contents of the source sub-
		record into the new subrecord.
*/
		recptr = newrec;

		addbyte( obj$c_gsd, &recptr );
		transfer_contents( &record[1], block->mask, &recptr );

/*		Our  new  subrecord is complete up to the argument section.  Now
		we add the minimum and maximum argument counts and a formal arg-
		ument descriptor for each argument.
*/
		arg_cnt = strlen( block->args );

		addbyte( block->min_args, &recptr );
		addbyte( arg_cnt, &recptr );

		for ( i = 0; i < arg_cnt; ++i ) {

			switch ( block->args[i] ) {

			case 'D':
			addbyte( arg$c_desc,    &recptr );
			break;

			case 'R':
			addbyte( arg$c_ref,     &recptr );
			break;

			case 'U':
			addbyte( arg$c_unknown, &recptr );
			break;

			case 'V':
			addbyte( arg$c_value,   &recptr );
			break;

			}

			addbyte( 0, &recptr );

		}

/*		The GSD record containing the single  GSD$C_(L)PRO(W)  subrecord
		is  complete.  After the record is written out we send a message
		to the screen indicating the  symbol's  defining  subrecord  has
		been modified.
*/
		write_record( newrec, recptr-newrec, rab );

		if ( LOGGING ) {
			lib$signal( MODOBJ$_MOD, 2, strlen( name ), name );
		}

	}
	else {

/*		The  subrecord  is  not  one  we're capable of modifying, or the
		symbol is not on our list.  We simply write the subrecord out.
*/
		write_record( record, 1 + length_gsd_subrecord( &record[1] ),
		              rab );

	}

	return;
}
/*============================================================================*/

/*============================================================================*/

	void	write_record ( record, reclen, rab )

/*	Writes a  record to a given file.  It is  assumed that all fields within
	the  specified RAB  are  set  up with  the exception  of  RAB$L_RBF  and
	RAB$W_RSZ.
*/
	ubyte	record[];
	ulong	reclen;
	struct	RAB *rab;
{
	ulong	status;

	void	lib$signal();
	ulong	sys$put();

/*----------------------------------------------------------------------------*/

	rab->rab$l_rbf = record;
	rab->rab$w_rsz = reclen;

	status = sys$put( rab );

	if ( (status & STS$M_SUCCESS) == 0 ) {
		lib$signal( MODOBJ$_OFE, 0, status, rab->rab$l_stv );
	}

	return;
}
/*============================================================================*/
